@@ -58,7 +58,7 @@
 
 #define DEFAULT_XML_BUF_SIZE	1024
 #define DISCONNECT_TIMER	2
-#define DISCOVERY_TIMER		2000
+#define DISCOVERY_TIMER		2
 
 struct btd_driver_data {
 	struct btd_device_driver *driver;
@@ -95,24 +95,17 @@ struct browse_req {
 	DBusConnection *conn;
 	DBusMessage *msg;
 	struct btd_device *device;
-	GSList *uuids;
-	GSList *uuids_added;
-	GSList *uuids_removed;
+	GSList *match_uuids;
+	GSList *profiles_added;
+	GSList *profiles_removed;
 	sdp_list_t *records;
 	int search_uuid;
 };
 
 static uint16_t uuid_list[] = {
-	PUBLIC_BROWSE_GROUP,
+	L2CAP_UUID,
 	PNP_INFO_SVCLASS_ID,
-	HID_SVCLASS_ID,
-	GENERIC_AUDIO_SVCLASS_ID,
-	ADVANCED_AUDIO_SVCLASS_ID,
-	AV_REMOTE_SVCLASS_ID,
-	PANU_SVCLASS_ID,
-	GN_SVCLASS_ID,
-	NAP_SVCLASS_ID,
-	SERIAL_PORT_SVCLASS_ID,
+	PUBLIC_BROWSE_GROUP,
 	0
 };
 
@@ -261,6 +254,13 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	ptr = adapter_get_path(adapter);
 	dict_append_entry(&dict, "Adapter", DBUS_TYPE_OBJECT_PATH, &ptr);
 
+	if (read_remote_eir(&src, &device->bdaddr, NULL) < 0)
+		boolean = TRUE;
+	else
+		boolean = FALSE;
+
+	dict_append_entry(&dict, "LegacyPairing", DBUS_TYPE_BOOLEAN, &boolean);
+
 	dbus_message_iter_close_container(&iter, &dict);
 
 	return reply;
@@ -402,6 +402,8 @@ static DBusMessage *discover_services(DBusConnection *conn,
 		if (bt_string2uuid(&uuid, pattern) < 0)
 			return invalid_args(msg);
 
+		sdp_uuid128_to_uuid(&uuid);
+
 		err = device_browse(device, conn, msg, &uuid, FALSE);
 		if (err < 0)
 			goto fail;
@@ -551,15 +553,14 @@ struct btd_device *device_create(DBusConnection *conn, struct btd_adapter *adapt
 {
 	gchar *address_up;
 	struct btd_device *device;
-	uint16_t dev_id = adapter_get_dev_id(adapter);
+	const gchar *adapter_path = adapter_get_path(adapter);
 
 	device = g_try_malloc0(sizeof(struct btd_device));
 	if (device == NULL)
 		return NULL;
 
 	address_up = g_ascii_strup(address, -1);
-	device->path = g_strdup_printf("%s/hci%d/dev_%s", "/org/bluez",
-							dev_id, address_up);
+	device->path = g_strdup_printf("%s/dev_%s", adapter_path, address_up);
 	g_strdelimit(device->path, ":", '_');
 	g_free(address_up);
 
@@ -607,50 +608,118 @@ gint device_address_cmp(struct btd_device *device, const gchar *address)
 	return strcasecmp(addr, address);
 }
 
-void device_probe_drivers(struct btd_device *device, GSList *uuids)
+static gboolean record_has_uuid(const sdp_record_t *rec,
+				const char *profile_uuid)
+{
+	sdp_list_t *pat;
+
+	for (pat = rec->pattern; pat != NULL; pat = pat->next) {
+		char *uuid;
+		int ret;
+
+		uuid = bt_uuid2string(pat->data);
+		if (!uuid)
+			continue;
+
+		ret = strcasecmp(uuid, profile_uuid);
+
+		g_free(uuid);
+
+		if (ret == 0)
+			return TRUE;
+	}
+
+	return FALSE;
+}
+
+static GSList *device_match_pattern(struct btd_device *device,
+					const char *match_uuid,
+					GSList *profiles)
+{
+	GSList *l, *uuids = NULL;
+
+	for (l = profiles; l; l = l->next) {
+		char *profile_uuid = l->data;
+		const sdp_record_t *rec;
+
+		rec = btd_device_get_record(device, profile_uuid);
+		if (!rec)
+			continue;
+
+		if (record_has_uuid(rec, match_uuid))
+			uuids = g_slist_append(uuids, profile_uuid);
+	}
+
+	return uuids;
+}
+
+static GSList *device_match_driver(struct btd_device *device,
+					struct btd_device_driver *driver,
+					GSList *profiles)
+{
+	const char **uuid;
+	GSList *uuids = NULL;
+
+	for (uuid = driver->uuids; *uuid; uuid++) {
+		GSList *match;
+
+		/* skip duplicated uuids */
+		if (g_slist_find_custom(uuids, *uuid,
+				(GCompareFunc) strcasecmp))
+			continue;
+
+		/* match profile driver */
+		match = g_slist_find_custom(profiles, *uuid,
+					(GCompareFunc) strcasecmp);
+		if (match) {
+			uuids = g_slist_append(uuids, match->data);
+			continue;
+		}
+
+		/* match pattern driver */
+		match = device_match_pattern(device, *uuid, profiles);
+		for (; match; match = match->next)
+			uuids = g_slist_append(uuids, match->data);
+	}
+
+	return uuids;
+}
+
+void device_probe_drivers(struct btd_device *device, GSList *profiles)
 {
 	GSList *list;
-	const char **uuid;
 	int err;
 
 	debug("Probe drivers for %s", device->path);
 
 	for (list = device_drivers; list; list = list->next) {
 		struct btd_device_driver *driver = list->data;
-		GSList *probe_uuids = NULL;
+		GSList *probe_uuids;
+		struct btd_driver_data *driver_data;
 
-		for (uuid = driver->uuids; *uuid; uuid++) {
-			GSList *match;
+		probe_uuids = device_match_driver(device, driver, profiles);
 
-			match = g_slist_find_custom(uuids, *uuid,
-							(GCompareFunc) strcasecmp);
-			if (!match)
-				continue;
+		if (!probe_uuids)
+			continue;
 
-			probe_uuids = g_slist_append(probe_uuids, match->data);
-		}
+		driver_data = g_new0(struct btd_driver_data, 1);
 
-		if (probe_uuids) {
-			struct btd_driver_data *driver_data = g_new0(struct btd_driver_data, 1);
+		err = driver->probe(device, probe_uuids);
+		if (err < 0) {
+			error("probe failed with driver %s for device %s",
+					driver->name, device->path);
 
-			err = driver->probe(device, probe_uuids);
-			if (err < 0) {
-				error("probe failed with driver %s for device %s",
-							driver->name, device->path);
-
-				g_free(driver_data);
-				g_slist_free(probe_uuids);
-				continue;
-			}
-
-			driver_data->driver = driver;
-			device->drivers = g_slist_append(device->drivers,
-								driver_data);
+			g_free(driver_data);
 			g_slist_free(probe_uuids);
+			continue;
 		}
+
+		driver_data->driver = driver;
+		device->drivers = g_slist_append(device->drivers, driver_data);
+		g_slist_free(probe_uuids);
 	}
 
-	for (list = uuids; list; list = list->next) {
+	for (list = profiles; list; list = list->next) {
 		GSList *l = g_slist_find_custom(device->uuids, list->data,
 							(GCompareFunc) strcasecmp);
 		if (l)
@@ -668,7 +737,7 @@ void device_probe_drivers(struct btd_device *device, GSList *uuids)
 	}
 }
 
-void device_remove_drivers(struct btd_device *device, GSList *uuids)
+static void device_remove_drivers(struct btd_device *device, GSList *uuids)
 {
 	struct btd_adapter *adapter = device_get_adapter(device);
 	GSList *list, *next;
@@ -698,23 +767,23 @@ void device_remove_drivers(struct btd_device *device, GSList *uuids)
 					(GCompareFunc) strcasecmp))
 				continue;
 
-			debug("UUID %s was removed from device %s", dstaddr);
+			debug("UUID %s was removed from device %s", *uuid, dstaddr);
 
 			driver->remove(device);
 			device->drivers = g_slist_remove(device->drivers,
 								driver_data);
-
 			g_free(driver_data);
 
-
 			rec = find_record_in_list(records, *uuid);
 			if (!rec)
-				continue;
+				break;
 
 			delete_record(srcaddr, dstaddr, rec->handle);
 
 			records = sdp_list_remove(records, rec);
 			sdp_record_free(rec);
+
+			break;
 		}
 	}
 
@@ -813,8 +882,7 @@ static void discover_device_reply(struct browse_req *req, sdp_list_t *recs)
 
 	dbus_message_iter_close_container(&iter, &dict);
 
-	dbus_connection_send(req->conn, reply, NULL);
-	dbus_message_unref(reply);
+	g_dbus_send_message(req->conn, reply);
 }
 
 static void services_changed(struct btd_device *device)
@@ -856,9 +924,8 @@ static void update_services(struct browse_req *req, sdp_list_t *recs)
 
 	for (seq = recs; seq; seq = seq->next) {
 		sdp_record_t *rec = (sdp_record_t *) seq->data;
-		sdp_buf_t pdu;
 		sdp_list_t *svcclass = NULL;
-		gchar *uuid_str;
+		gchar *profile_uuid;
 		GSList *l;
 
 		if (!rec)
@@ -868,13 +935,13 @@ static void update_services(struct browse_req *req, sdp_list_t *recs)
 			continue;
 
 		/* Extract the first element and skip the remainning */
-		uuid_str = bt_uuid2string(svcclass->data);
-		if (!uuid_str) {
+		profile_uuid = bt_uuid2string(svcclass->data);
+		if (!profile_uuid) {
 			sdp_list_free(svcclass, free);
 			continue;
 		}
 
-		if (!strcasecmp(uuid_str, PNP_UUID)) {
+		if (!strcasecmp(profile_uuid, PNP_UUID)) {
 			uint16_t source, vendor, product, version;
 			sdp_data_t *pdlist;
 
@@ -895,19 +962,9 @@ static void update_services(struct browse_req *req, sdp_list_t *recs)
 						vendor, product, version);
 		}
 
-		/* Driver UUID found */
-		l = g_slist_find_custom(req->uuids, uuid_str,
-				(GCompareFunc) strcasecmp);
-		if (l) {
-			char *uuid = l->data;
-
-			req->uuids = g_slist_remove(req->uuids, uuid);
-			g_free(uuid);
-		}
-
 		/* Check for duplicates */
 		if (sdp_list_find(req->records, rec, rec_cmp)) {
-			g_free(uuid_str);
+			g_free(profile_uuid);
 			sdp_list_free(svcclass, free);
 			continue;
 		}
@@ -915,25 +972,19 @@ static void update_services(struct browse_req *req, sdp_list_t *recs)
 		store_record(srcaddr, dstaddr, rec);
 
 		/* Copy record */
-		if (sdp_gen_record_pdu(rec, &pdu) == 0) {
-			sdp_record_t *record;
-			int scanned;
+		req->records = sdp_list_append(req->records, sdp_copy_record(rec));
 
-			record = sdp_extract_pdu(pdu.data, pdu.data_size,
-						&scanned);
-			free(pdu.data);
-			req->records = sdp_list_append(req->records, record);
-		}
-
-		l = g_slist_find_custom(device->uuids, uuid_str,
+		l = g_slist_find_custom(device->uuids, profile_uuid,
 							(GCompareFunc) strcmp);
 		if (!l)
-			req->uuids_added = g_slist_append(req->uuids_added,
-								uuid_str);
+			req->profiles_added =
+					g_slist_append(req->profiles_added,
+							profile_uuid);
 		else {
-			req->uuids_removed = g_slist_remove(req->uuids_removed,
-								l->data);
-			g_free(uuid_str);
+			req->profiles_removed =
+					g_slist_remove(req->profiles_removed,
+							l->data);
+			g_free(profile_uuid);
 		}
 
 		sdp_list_free(svcclass, free);
@@ -975,9 +1026,9 @@ static void browse_req_free(struct browse_req *req)
 		dbus_message_unref(req->msg);
 	if (req->conn)
 		dbus_connection_unref(req->conn);
-	g_slist_foreach(req->uuids_added, (GFunc) g_free, NULL);
-	g_slist_free(req->uuids_added);
-	g_slist_free(req->uuids_removed);
+	g_slist_foreach(req->profiles_added, (GFunc) g_free, NULL);
+	g_slist_free(req->profiles_added);
+	g_slist_free(req->profiles_removed);
 	if (req->records)
 		sdp_list_free(req->records, (sdp_free_func_t) sdp_record_free);
 	g_free(req);
@@ -997,7 +1048,7 @@ static void search_cb(sdp_list_t *recs, int err, gpointer user_data)
 
 	update_services(req, recs);
 
-	if (!req->uuids_added && !req->uuids_removed) {
+	if (!req->profiles_added && !req->profiles_removed) {
 		debug("%s: No service update", device->path);
 		goto proceed;
 	}
@@ -1009,12 +1060,12 @@ static void search_cb(sdp_list_t *recs, int err, gpointer user_data)
 	}
 
 	/* Probe matching drivers for services added */
-	if (req->uuids_added)
-		device_probe_drivers(device, req->uuids_added);
+	if (req->profiles_added)
+		device_probe_drivers(device, req->profiles_added);
 
 	/* Remove drivers for services removed */
-	if (req->uuids_removed)
-		device_remove_drivers(device, req->uuids_removed);
+	if (req->profiles_removed)
+		device_remove_drivers(device, req->profiles_removed);
 
 	/* Propagate services changes */
 	services_changed(req->device);
@@ -1023,25 +1074,15 @@ proceed:
 	/* Store the device's profiles in the filesystem */
 	store_profiles(device);
 
-	if (req->msg) {
-		if (dbus_message_is_method_call(req->msg, DEVICE_INTERFACE,
-						"DiscoverServices")) {
-			discover_device_reply(req, req->records);
-			goto cleanup;
-		}
-
-		g_dbus_emit_signal(req->conn, dbus_message_get_path(req->msg),
-				ADAPTER_INTERFACE, "DeviceCreated",
-				DBUS_TYPE_OBJECT_PATH, &device->path,
-				DBUS_TYPE_INVALID);
-	}
-
-	/* Update device list */
-	adapter_update_devices(device->adapter);
-
 	if (!req->msg)
 		goto cleanup;
 
+	if (dbus_message_is_method_call(req->msg, DEVICE_INTERFACE,
+					"DiscoverServices")) {
+		discover_device_reply(req, req->records);
+		goto cleanup;
+	}
+
 	/* Reply create device request */
 	reply = dbus_message_new_method_return(req->msg);
 	if (!reply)
@@ -1050,8 +1091,7 @@ proceed:
 	dbus_message_append_args(reply, DBUS_TYPE_OBJECT_PATH, &device->path,
 							DBUS_TYPE_INVALID);
 
-	dbus_connection_send(req->conn, reply, NULL);
-	dbus_message_unref(reply);
+	g_dbus_send_message(req->conn, reply);
 
 cleanup:
 	browse_req_free(req);
@@ -1065,9 +1105,9 @@ static void browse_cb(sdp_list_t *recs, int err, gpointer user_data)
 	bdaddr_t src;
 	uuid_t uuid;
 
-	/* If we have a valid response and req->search_uuid == 1, then
-	   public browsing was successful -- we don't need any more */
-	if (err < 0 || (req->search_uuid == 1 && recs))
+	/* If we have a valid response and req->search_uuid == 2, then
+	   L2CAP UUID & PNP searching was successful -- we are done */
+	if (err < 0 || (req->search_uuid == 2 && req->records))
 		goto done;
 
 	update_services(req, recs);
@@ -1082,26 +1122,6 @@ static void browse_cb(sdp_list_t *recs, int err, gpointer user_data)
 		return;
 	}
 
-	/* Search for drivers uuids */
-	if (req->uuids) {
-		char *uuid_str = req->uuids->data;
-
-		bt_string2uuid(&uuid, uuid_str);
-		req->uuids = g_slist_remove(req->uuids, uuid_str);
-		g_free(uuid_str);
-		bt_search_service(&src, &device->bdaddr, &uuid,
-						browse_cb, user_data, NULL);
-		return;
-	}
-
-	/* Search for l2cap uuid */
-	if (!req->records) {
-		sdp_uuid16_create(&uuid, L2CAP_UUID);
-		bt_search_service(&src, &device->bdaddr, &uuid,
-						search_cb, user_data, NULL);
-		return;
-	}
-
 done:
 	search_cb(recs, err, user_data);
 }
@@ -1110,37 +1130,6 @@ static void init_browse(struct browse_req *req, gboolean reverse)
 {
 	GSList *l;
 
-	for (l = device_drivers; l; l = l->next) {
-		struct btd_device_driver *driver = l->data;
-		int i;
-
-		for (i = 0; driver->uuids[i]; i++) {
-			char *uuid;
-			int j;
-
-			/* Eliminate duplicates of UUIDs in uuid_list[]... */
-			if (strlen(driver->uuids[i]) == 36 && 
-			    !strncmp(driver->uuids[i], "0000", 4) &&
-			    !strcasecmp(driver->uuids[i] + 8, 
-					"-0000-1000-8000-00805F9B34FB")) {
-				uint16_t uuid16 = strtol(driver->uuids[i],
-							 NULL, 16);
-				for (j = 0; uuid_list[j]; j++) {
-					if (uuid16 == uuid_list[j])
-						continue;
-				}
-
-			}
-			/* ... and of UUIDs another driver already asked for */
-			if (g_slist_find_custom(req->uuids, driver->uuids[i],
-					(GCompareFunc) strcasecmp))
-				continue;
-
-			uuid = g_strdup(driver->uuids[i]);
-			req->uuids = g_slist_append(req->uuids, uuid);
-		}
-	}
-
 	/* If we are doing reverse-SDP don't try to detect removed profiles
 	 * since some devices hide their service records while they are
 	 * connected
@@ -1149,7 +1138,7 @@ static void init_browse(struct browse_req *req, gboolean reverse)
 		return;
 
 	for (l = req->device->uuids; l; l = l->next)
-		req->uuids_removed = g_slist_append(req->uuids_removed,
+		req->profiles_removed = g_slist_append(req->profiles_removed,
 						l->data);
 }
 
@@ -1169,10 +1158,9 @@ int device_browse(struct btd_device *device, DBusConnection *conn,
 	adapter_get_address(adapter, &src);
 
 	req = g_new0(struct browse_req, 1);
-	if (conn && msg) {
-		req->conn = dbus_connection_ref(conn);
-		req->msg = dbus_message_ref(msg);
-	}
+
+	req->conn = conn ? dbus_connection_ref(conn) : get_dbus_connection();
+
 	req->device = device;
 
 	if (search) {
@@ -1187,6 +1175,7 @@ int device_browse(struct btd_device *device, DBusConnection *conn,
 	device->discov_active = 1;
 
 	if (msg) {
+		req->msg = dbus_message_ref(msg);
 		device->discov_requestor = g_strdup(dbus_message_get_sender(msg));
 		/* Track the request owner to cancel it
 		 * automatically if the owner exits */
@@ -1320,7 +1309,7 @@ int device_set_paired(DBusConnection *conn, struct btd_device *device,
 	/* If we are not initiators and there is no currently active discovery
 	 * or discovery timer, set the discovery timer */
 	if (!device->discov_active && !device->discov_timer)
-		device->discov_timer = g_timeout_add(DISCOVERY_TIMER,
+		device->discov_timer = g_timeout_add_seconds(DISCOVERY_TIMER,
 							start_discovery,
 							device);
 
